#include "elements.h"

static unsigned int GetBasicTypeBytes (BasicType bt)
{
    unsigned int ret_val = 0;
    switch (bt)
    {
    case (BT_FLOAT) : ret_val = sizeof (float); break;
    case (BT_CHAR)  : ret_val = sizeof (char); break;
    case (BT_UINT)  : ret_val = sizeof (int); break;
    case (BT_USHORT): ret_val = sizeof (unsigned short int); break;
    }
    return ret_val;
}



ElementArray::ElementArray (ElementType et, BasicType bt, unsigned int bt_count)
{
    element_type = et;
    basic_type = bt;
    basic_type_count = bt_count;

    element_size = basic_type_count * GetBasicTypeBytes (bt);
    data = NULL;
    data_size_in_bytes = 0;
    elements_count = 0;
}

ElementArray::~ElementArray ()
{
    Release ();
}


unsigned int ElementArray::GetBasicTypeCount () const
{
    return basic_type_count;
}

unsigned int ElementArray::GetElementsCount () const
{
    return elements_count;
}

unsigned int ElementArray::GetSingleElementSize () const
{
    return element_size;
}


void *ElementArray::Create (unsigned int element_count)
{
    Release ();
    this->elements_count = element_count;
    data_size_in_bytes = elements_count * element_size;
	try
	{
		data = new char [data_size_in_bytes];
	}
	catch (...)
	{
		data = NULL;
	}

    return data;
}

void *ElementArray::GetData ()
{
    return data;
}

void ElementArray::Release ()
{
    delete [] data;
    data = NULL;
    data_size_in_bytes = 0;
    elements_count = 0;
}


bool ElementArray::LoadFromFP (VFS_FILE *fp)
{
    try
    {
        if (data == NULL)
        {
            throw (int) 0;
        }

        size_t bytes_read;
        bytes_read = vfs_read (data, 1, data_size_in_bytes,  fp);
        if (bytes_read != data_size_in_bytes)
        {
            throw (int) 1;
        }
    }
    catch (...)
    {
        return false;
    }
    return true;
}

bool ElementArray::WriteToFP (VFS_FILE *fp)
{
	if (data == NULL)
	{
		return true;
	}
	size_t bytes_write = vfs_write (data, 1, data_size_in_bytes, fp);
	if (bytes_write != data_size_in_bytes)
	{
		return false;
	}
	return true;
}



bool ElementArray::MapWithStride (void *dest, unsigned int stride)
{
    if (data == NULL)
    {
        return false;
    }
    char *dest_ptr = (char *) dest;
    char *src_ptr = data;

    for (unsigned int i = 0; i < elements_count; i ++)
    {
        memcpy (dest_ptr, src_ptr, element_size);

        src_ptr += element_size;
        dest_ptr += stride;
    }
    

    return true;
}


VertexArray::VertexArray () : ElementArray (ET_VERTEX, BT_FLOAT, 3)
{
}

VertexArray::~VertexArray ()
{
}



NormalArray::NormalArray () : ElementArray (ET_NORMAL, BT_FLOAT, 3)
{
}

NormalArray::~NormalArray ()
{
}


TexCoordArray::TexCoordArray (unsigned int tex_coords) : ElementArray (ET_TEXCOORD, BT_FLOAT, tex_coords)
{
}

TexCoordArray::~TexCoordArray ()
{
}


TangentArray::TangentArray () : ElementArray (ET_TANGENT, BT_FLOAT, 3)
{
}

TangentArray::~TangentArray ()
{
}


ColorArray::ColorArray () : ElementArray (ET_COLOR, BT_UINT, 1)
{
}

ColorArray::~ColorArray ()
{
}


IndexArray::IndexArray () : ElementArray (ET_INDEX, BT_UINT, 1)
{
}

IndexArray::~IndexArray ()
{
}

BoneIndexArray::BoneIndexArray() : ElementArray(ET_BONE_INDEX, BT_CHAR, 4)
{
}

BoneIndexArray::~BoneIndexArray()
{
}

BoneWeightArray::BoneWeightArray() : ElementArray(ET_BONE_WEIGHT, BT_FLOAT, 3)
{
}

BoneWeightArray::~BoneWeightArray()
{
}
