#include <algorithm>
#include <gpuprog.h>
#include "meshman.h"

//TODO : check correct free of memory and objects


static const unsigned int max_vertex_count = 500000;
static const unsigned int max_index_count = 500000;


bool MeshInfo::operator<(const MeshInfo& info) const
{
	//this < info ?

	if (vf.GetOrdinal() < info.vf.GetOrdinal())
	{
		return true;
	}
	if (vf.GetOrdinal() > info.vf.GetOrdinal())
	{
		return false;
	}
	if (vf.GetOrdinal() == info.vf.GetOrdinal())
	{
		if (mat_id < info.mat_id)
		{
			return true;
		}
	}
	return false;
}


MeshBuffer::MeshBuffer (VertexFormat &vf)
{
	buffer_vertex_format = vf;

	hw_vbuffer = 0;
	hw_ibuffer = 0;

	total_vertices = 0;
	total_indices = 0;
}

MeshBuffer::~MeshBuffer ()
{
	Release ();
}

void MeshBuffer::Release ()
{
	if (hw_vbuffer != 0)
	{
		DeleteVertexBuffer (hw_vbuffer);
		hw_vbuffer = 0;
	}
	if (hw_ibuffer != 0)
	{
		DeleteIndexBuffer (hw_ibuffer);
		hw_ibuffer = 0;
	}

	for (unsigned int i = 0; i < mesh_list.size(); i++)
	{
		mesh_list[i]->rg_info_ptr->ib_id = 0;
		mesh_list[i]->rg_info_ptr->vb_id = 0;
	}
	mesh_list.clear();

	total_vertices = 0;
	total_indices = 0;
}


bool MeshBuffer::AddMesh (MeshInfo *info)
{
	if (info->vf != buffer_vertex_format)
	{
		return false;
	}
	if (total_vertices + info->geom->GetVertexCount() >= max_vertex_count)
	{
		return false;
	}
	if (total_indices + info->geom->GetIndexCount() >= max_index_count)
	{
		return false;
	}

	info->rg_info_ptr->start_vertex = total_vertices;
	info->rg_info_ptr->vertex_count = info->geom->GetVertexCount();
	info->rg_info_ptr->start_index = total_indices;
	info->rg_info_ptr->index_count = info->geom->GetIndexCount();
	info->rg_info_ptr->primitive_count = info->geom->GetPrimitiveCount ();
	info->rg_info_ptr->vb_id = 0;
	info->rg_info_ptr->ib_id = 0;

	mesh_list.push_back (info);

	total_vertices += info->geom->GetVertexCount();
	total_indices += info->geom->GetIndexCount ();

	return true;
}

bool MeshBuffer::Build ()
{
	hw_vbuffer = CreateVertexBuffer (buffer_vertex_format, total_vertices, false);
	if (hw_vbuffer == 0)
	{
		return false;
	}

	unsigned char *vptr = (unsigned char *)LockVertexBuffer (hw_vbuffer, 0, 0);
	if (vptr == NULL)
	{
		return false;
	}

	unsigned int vertex_size = buffer_vertex_format.GetSize();

	for (unsigned int i = 0; i < mesh_list.size(); i++)
	{
		if (mesh_list[i]->geom->MapGeometry ((void*)&vptr[mesh_list[i]->rg_info_ptr->start_vertex*vertex_size], buffer_vertex_format) == false)
		{
			UnlockVertexBuffer (hw_vbuffer);
			return false;
		}
		mesh_list[i]->rg_info_ptr->vb_id = hw_vbuffer;
	}
	UnlockVertexBuffer (hw_vbuffer);


	hw_ibuffer = CreateIndexBuffer (total_indices, false);
	if (hw_ibuffer == 0)
	{
		return false;
	}
	unsigned int *iptr = (unsigned int*) LockIndexBuffer (hw_ibuffer, 0, 0);
	if (iptr == NULL)
	{
		return false;
	}
	for (unsigned int i = 0; i < mesh_list.size(); i++)
	{
		if (mesh_list[i]->geom->MapIndex ((void*) &iptr[mesh_list[i]->rg_info_ptr->start_index], mesh_list[i]->rg_info_ptr->start_vertex) == false)
		{
			UnlockIndexBuffer (hw_ibuffer);
			return false;
		}
		mesh_list[i]->rg_info_ptr->ib_id = hw_ibuffer;
	}
	UnlockIndexBuffer (hw_ibuffer);

	return true;
}







MeshManager::MeshManager ()
{
}

MeshManager::~MeshManager ()
{
	Release ();
}

void MeshManager::Release ()
{
	ClearMeshBuffers ();
	mesh_info.clear ();
}

void MeshManager::ClearMeshBuffers ()
{
	for (unsigned int i = 0; i < mesh_buffers.size(); i++)
	{
		delete mesh_buffers[i];
	}
	mesh_buffers.clear();
}


bool MeshManager::AddGeometry (Geometry *geom, unsigned int mat_id, const VertexFormat &vf, RGInfo *rg_info)
{
	if ((geom->GetVertexArray () == NULL) || (geom->GetIndexArray() == NULL) || (rg_info == NULL))
	{
		return false;
	}

	MeshInfo info;

	info.geom = geom;
	info.mat_id = mat_id;
	info.vf = vf;
	info.rg_info_ptr = rg_info;

	mesh_info.push_back (info);
	return true;
}


bool MeshManager::Build ()
{
	ClearMeshBuffers ();

	//sort by vertex format and material
	sort(mesh_info.begin(), mesh_info.end());

	for (unsigned int i = 0; i < mesh_info.size(); i++)
	{
		bool added = false;
		for (unsigned int j = 0; j < mesh_buffers.size(); j++)
		{
			MeshBuffer *mb = mesh_buffers[j];
			if (mb->AddMesh (&mesh_info[i]))
			{
				added = true;
				break;
			}
		}

		if (added == false)
		{
			MeshBuffer *mb = new MeshBuffer (mesh_info[i].vf);
			mesh_buffers.push_back (mb);
			if (mb->AddMesh (&mesh_info[i]) == false)
			{
				return false;
			}
		}
	}

	for (unsigned int i = 0; i < mesh_buffers.size(); i++)
	{
		if (mesh_buffers[i]->Build () == false)
		{
			return false;
		}
	}

	return true;
}

unsigned int MeshManager::GetMeshCount ()
{
	return (unsigned int) mesh_buffers.size();
}

